#
# Copyright (C) Fujitsu Technology Solutions 2009
# All Rights Reserved.
#
# Copyright (C) Fujitsu Siemens Computers 2006 2007 2008
# All Rights Reserved.
#
# $Id: macaddr.awk,v 1.5 2006-12-07 17:42:57+01 richter Exp richter $
#
# NAME
#	macaddr.awk - map "virtual" to "real" MAC addresses on VMware ESX 3.0
#
# SYNOPSIS
#	gawk [-vTRACE=1] [-vTEST=1] -f <thisScript> /etc/vmware/esx.conf
#
# DESCRIPTION
#	For VMware ESX 3.0 construct a table for "virtual" to "real" MAC address
#	mapping. This script assumes that it is only called in appropriate situations.
#	There are no further checks. The TRACE and TEST variables are reserved
#	for diagnostic purposes.
#
#	The conceptual idea is as follows:
#	1. each vswif<n> is uniquely associated with a portgroup noted in its
#	   ifcfg-vswif<n> network-script
#	2. one or more portgroups are assigned with one or more vmnic<m>
#	   used as their uplinks
#	3. a vswif<n> can be associated by its portgroup with one of the
#	   (many) vmnic<m> uplinks for this portgroup
#
#	Example output table (the header is not part of the table):
#
#	vswifn   virt MAC Addr     real MAC Addr   NIC
#	vswif0 00:50:56:42:5e:92 00:30:05:2F:A4:8C vmnic0
#
#	This table is generated from information from stdin together with
#	the output of ifconfig as follows:
#
#	1. From stdin filter lines to get information about "vswif<n>"
#		/net/vswifs/vswif<n> = "/etc/sysconfig/network-scripts/ifcfg-vswif<n>"
#
#	2. From such a "ifcfg-vswif<n>" file like
#		# Broadcom Corporation|NetXtreme BCM5704 Gigabit Ethernet
#		DEVICE=<device>
#		BOOTPROTO=dhcp
#		ONBOOT=yes
#		PORTGROUP=<portgroup>
#		MACADDR=<macaddr>
#	  filter DEVICE, PORTGROUP, and MACADDR values into array
#		vswifs[<n>, "DEVICE"] = <DEVICE>
#		vswifs[<n>, "PORTGROUP"] = <portgroup>
#		vswifs[<n>, "MACADDR"]> = <macaddr>
#
#	3. From stdin filter lines
#		/net/vswitch/child[<k>]/portgroup/child[<l>]/uid = <portgroup>
#		/net/vswitch/child[<i>]/uplinks/child[<j>]/pnic = <vmnic>
#	   into arrays
#		portgroups[<k>, <l>] = <portgroup>
#		uplinks[<i>, <j>] = <vmnic>
#
#	4. For each vswif<n> locate its portgroup in portgroups[<k>, <l>]
#	   and associate with vmnic<m> from uplinks[<k>, <l>], if it exists
#	   or the first (gawk array enumeration!) entry uplinks[<k>, *].
#
#	5. retrieve HWaddr of this vmnic<m> from ifconfig output
#
#	6. print table line
#
#	Example input:
#
#	/net/vswifs/vswif0 = "/etc/sysconfig/network-scripts/ifcfg-vswif0"
#	/net/vswifs/vswif1 = "/etc/sysconfig/network-scripts/ifcfg-vswif1"
#
#	/net/vswitch/child[0000]/portgroup/child[0000]/uid = "portgroup0"
#	/net/vswitch/child[0000]/portgroup/child[0001]/uid = "portgroup1"
#	/net/vswitch/child[0000]/uplinks/child[0000]/pnic = "vmnic0"
#
#	/net/vswitch/child[0001]/portgroup/child[0000]/uid = "portgroup9"
#	/net/vswitch/child[0001]/portgroup/child[0001]/uid = "portgroup13"
#	/net/vswitch/child[0001]/portgroup/child[0002]/uid = "portgroup14"
#	/net/vswitch/child[0001]/portgroup/child[0003]/uid = "portgroup19"
#	/net/vswitch/child[0001]/uplinks/child[0000]/pnic = "vmnic1"
#
#	/net/vswitch/child[0002]/portgroup/child[0000]/uid = "portgroup20"
#	/net/vswitch/child[0002]/uplinks/child[0000]/pnic = "vmnic2"
#	/net/vswitch/child[0002]/uplinks/child[0001]/pnic = "vmnic3"
#
#	example "$(ifconfig)" output
#
#	vmnic0    Link encap:Ethernet  HWaddr 00:30:05:2F:A4:8C
#	vmnic1    Link encap:Ethernet  HWaddr 00:30:05:2F:A4:8D
#	vswif0    Link encap:Ethernet  HWaddr 00:50:56:42:5E:92
#
#	The algorithm on VMware ESX from 4.0 is similar, but we have also
#	to take care of the teampolicy (see below).
#


BEGIN {
	vswifNr		= 0
	vmnicNr		= 0
	children	= 0

	# SVNickName might be e.g, ESX3, ESX3.5, ESX4, ESX4.0, ESX4.1, ESX5, etc.
	SVNickName	= ENVIRON["SV_NICKNAME"]

	ESXMajorVersion	= substr(SVNickName, 4, 1)		# skip "ESX"
	if (ESXMajorVersion == "")
		ESXMajorVersion = "3"

	if (substr(SVNickName, 5, 1) == ".")
	{
		ESXMinorVersion	= substr(SVNickName, 6, 1)	# skip "ESX", Major, and "."
	}
	else
	{
		ESXMinorVersion	= substr(SVNickName, 5, 1)	# skip "ESX" and Major
	}
	

	if (ESXMinorVersion == "")
		ESXMinorVersion = "0"

	ESXVersion = ESXMajorVersion ESXMinorVersion

	if (TRACE) {
		print "\tSVNickName     = " SVNickName
		print "\tESXVersion     = " ESXVersion
	}
}

# ---------------------------------------------------------------------
# From each /net/vswifs/vswif<n> = "/etc/sysconfig/network-scripts/ifcfg-vswif<n>"
# get the PORTGROUP and (virtual) MACADDR from the corresponding
# /etc/sysconfig/network-scripts/ifcfg-vswif<n>

$1 ~ /\/net\/vswifs\/vswif[[:digit:]]+/ {

	# Get vswif<n> number from the quoted 3rd argument "....<n>".

	ifcfgFile = trimquotes($3)
	if (TRACE)
		print "ifcfgFile = " ifcfgFile

	if (match(ifcfgFile, /[[:digit:]]+$/)) {
		vswifNr = substr(ifcfgFile, RSTART)
		if (TRACE)
			print "\tvswifNr:   " vswifNr
	}

	if (TEST)
		ifcfgFile = "./vswif" vswifNr
	else
		ifcfgFile = "/etc/sysconfig/network-scripts/ifcfg-vswif" vswifNr

	# Remember DEVICE, PORTGROUP and MACADDR from the ifcfg file
	# for association. Line format is:
	#	DEVICE=vswif0

	while ((getline <ifcfgFile) > 0) {
		if (split($0, arr, "=") == 2)
			vswifs[vswifNr, arr[1]] = trimquotes(arr[2])
	}
	close(ifcfgFile)
	if (TRACE) {
		print "\tDEVICE        :   " vswifs[vswifNr, "DEVICE"]
		print "\tHOTPLUG       :   " vswifs[vswifNr, "HOTPLUG"]
		print "\tMACADDR (virt):   " vswifs[vswifNr, "MACADDR"]
		print "\tONBOOT        :   " vswifs[vswifNr, "ONBOOT"]
		print "\tPORTGROUP     :   " vswifs[vswifNr, "PORTGROUP"]
		print "\tBOOTPROTO     :   " vswifs[vswifNr, "BOOTPROTO"]
		print "\tBROADCAST     :   " vswifs[vswifNr, "BROADCAST"]
		print "\tDHCPV6C       :   " vswifs[vswifNr, "DHCPV6C"]
		print "\tIPADDR        :   " vswifs[vswifNr, "IPADDR"]
		print "\tIPV6INIT      :   " vswifs[vswifNr, "IPV6INIT"]
		print "\tIPV6_AUTOCONF :   " vswifs[vswifNr, "IPV6_AUTOCONF"]
		print "\tNETMASK       :   " vswifs[vswifNr, "NETMASK"]
	}
	next
}

# ---------------------------------------------------------------------
# ---- OLD - STRUFF ----: ==> OBSOLETE
# This rule is only relevant on ESX versions less than ESX 3.5 (i.e. only for ESX 3.0)
# This rule is obsolete on ESX from ESX 3.5
# On EXS == 3.0 (and below)
# 	Frome each:
#		/net/vswitch/child[<k>]/portgroup/child[<l>]/uid = "portgroup<n>"
#	collect and remember from the (<k>,<l>) pair
#		portgroups[<k>, <l>] = "portgroup<n>"

$1 ~ /\/net\/vswitch\/child\[[[:digit:]]+\]\/portgroup\/child\[[[:digit:]]+\]\/uid/ {

	temp = $3
	if ($4 != "") temp = temp " " $4
	if ($5 != "") temp = temp " " $5
	if ($6 != "") temp = temp " " $6
	if ($7 != "") temp = temp " " $7

	if (TRACE == 2) {
		print "On ESX " ESXVersion ": /net/vswitch/child[<k>]/portgroup/child[<l>]/uid = " temp 
	}

	if (match($1, /child\[([[:digit:]]+)\].*child\[([[:digit:]]+)\]/, arr)) {
		k = arr[1] + 0
		l = arr[2] + 0
 
		portgroups[k, l] = trimquotes(temp) 
	}
	next
}

# ---------------------------------------------------------------------
# ---- NEW - STRUFF ----
# This rule is only relevant on ESX versions from ESX 3.5
# On EXS  >= 3.5
# 	Frome each:
#		/net/vswitch/child[<k>]/portgroup/child[<l>]/name = "portgroup<n>"
#	collect and remember from the (<k>,<l>) pair
#		portgroups[<k>, <l>] = "portgroup<n>"

$1 ~ /\/net\/vswitch\/child\[[[:digit:]]+\]\/portgroup\/child\[[[:digit:]]+\]\/name/ {

	temp = $3
	if ($4 != "") temp = temp " " $4
	if ($5 != "") temp = temp " " $5
	if ($6 != "") temp = temp " " $6
	if ($7 != "") temp = temp " " $7

	if (TRACE == 2) {
		print "On ESX " ESXVersion ": /net/vswitch/child[<k>]/portgroup/child[<l>]/name =" temp 
	}

	if (match($1, /child\[([[:digit:]]+)\].*child\[([[:digit:]]+)\]/, arr)) {
		k = arr[1] + 0
		l = arr[2] + 0

		portgroups[k, l] = trimquotes(temp)
	}
	next
} 

# ---------------------------------------------------------------------
# ---- NEW - STRUFF ----
# This rule is only relevant on ESX versions from ESX 4.0
# On EXS  >= 4.0
# 	Frome each:
#		/net/vswitch/child[<i>]/portgroup/child[<j>]/teamPolicy/uplinks[<k>]/pnic = "vmnic<m>"
#	collect and remember from the (<i>,<j>, <k>)
#		PortGroupUplinks[<i>, <j>, <k>] = "vmnic<m>"

$1 ~ /\/net\/vswitch\/child\[[[:digit:]]+\]\/portgroup\/child\[[[:digit:]]+\]\/teamPolicy\/uplinks\[[[:digit:]]+\]\/pnic/ {

	if (TRACE == 2) {
		print "On ESX " ESXVersion ": /net/vswitch/child[<i>]/portgroup/child[<j>]/teamPolicy/uplinks[<k>]/pnic =" $3 
	}

	if (match($1, /child\[([[:digit:]]+)\].*child\[([[:digit:]]+)\].*uplinks\[([[:digit:]]+)\]/, arr)) {
		i = arr[1] + 0
		j = arr[2] + 0
		k = arr[3] + 0

		PortGroupUplinks[i, j , k ] = trimquotes($3)
	}
	next
}

# ---------------------------------------------------------------------
# ---- NEW - STRUFF ----
# This rule is only relevant on ESX versions from ESX 4.0
# On EXS  >= 4.0
# 	Frome each:
#		/net/vswitch/child[<i>]/teamPolicy/uplinks[<j>]/pnic = "vmnic<m>"
#	collect and remember from the (<i>,<j>) pair
#		VswitchUplinks[<i>, <j>] = "vmnic<m>"

$1 ~ /\/net\/vswitch\/child\[[[:digit:]]+\]\/teamPolicy\/uplinks\[[[:digit:]]+\]\/pnic/ {

	if (TRACE == 2) {
		print "On ESX " ESXVersion ": /net/vswitch/child[<i>]/teamPolicy/uplinks[<j>]/pnic =" $3 
	}

	if (match($1, /child\[([[:digit:]]+)\].*uplinks\[([[:digit:]]+)\]/, arr)) {
		i = arr[1] + 0
		j = arr[2] + 0

		VswitchUplinks[i, j] = trimquotes($3)
	}
	next
}

# ---------------------------------------------------------------------
# ---- OLD - STRUFF ----: ==> OBSOLETE
# This rule might be obsolete now
# On EXS  < 4.0
# 	Frome each:
#		/net/vswitch/child[<i>]/uplinks[<j>]/pnic = "vmnic<m>"
#	collect and remember from the (<i>,<j>) pair
#		uplinks[<i>, <j>] = "vmnic<m>"

$1 ~ /\/net\/vswitch\/child\[[[:digit:]]+\]\/uplinks\/child\[[[:digit:]]+\]\/pnic/ {

	if (TRACE == 2) {
		print "On ESX " ESXVersion ": /net/vswitch/child[<i>]/uplinks/child[<j>]/pnic =" $3 
	}

	if (match($1, /child\[([[:digit:]]+)\].*child\[([[:digit:]]+)\]/, arr)) {
		i = arr[1] + 0
		j = arr[2] + 0

		uplinks[i, j] = trimquotes($3)
	}
	next}


# ---------------------------------------------------------------------
# ---- OLD - STRUFF ----: ==> OBSOLETE
# This rule might be obsolete now
# 	Frome each:
#		/net/dvswitch/child[<i>]/uplinks/child[<j>]/pnic = "vmnic<m>"
#	collect and remember from the (<i>,<j>) pair
#		uplinks[<i>, <j>] = "vmnic<m>"

$1 ~ /\/net\/dvswitch\/child\[[[:digit:]]+\]\/uplinks\/child\[[[:digit:]]+\]\/pnic/ {

	if (TRACE == 2) {
		print "On ESX " ESXVersion ": /net/dvswitch/child[<i>]/uplinks/child[<j>]/pnic =" $3 
	}

	if (match($1, /child\[([[:digit:]]+)\].*child\[([[:digit:]]+)\]/, arr)) {
		i = arr[1] + 0
		j = arr[2] + 0

		uplinks[i, j] = trimquotes($3)
	}
	next
}

# ---------------------------------------------------------------------
function trimquotes(s,	c) {

	c = substr(s, 1, 1)
	if (c != "\"" && c != "'")
		return (s)
	return (substr(s, 2, length(s) - 2))
}

# ---------------------------------------------------------------------

END {

	if (TRACE) {
		for (j in portgroups) {
		
			split (j, idxj, SUBSEP)
			print "portgroups["idxj[1]", "idxj[2]"] = " portgroups[j]
		}

		for (j in PortGroupUplinks) {
		
			split (j, idxj, SUBSEP)
			print "PortGroupUplinks["idxj[1]", "idxj[2]", "idxj[3]"] = " PortGroupUplinks[j]
		}

		for (j in VswitchUplinks) {
		
			split (j, idxj, SUBSEP)
			print "VswitchUplinks["idxj[1]", "idxj[2]"] = " VswitchUplinks[j]
		}

		for (j in uplinks) {
		
			split (j, idxj, SUBSEP)
			print "uplinks["idxj[1]", "idxj[2]"] = " uplinks[j]
		}
	}

	# Catch all HWaddr from ifconfig
	#
	#	vmnic0    Link encap:Ethernet  HWaddr 00:30:05:2F:A4:8C
	#	vmnic1    Link encap:Ethernet  HWaddr 00:30:05:2F:A4:8D
	#	vswif0    Link encap:Ethernet  HWaddr 00:50:56:42:5E:92

	if (TEST) {
		ifconfig = "./ifconfig"
		while ((getline <ifconfig) > 0) {
			if ($1 ~ /^vmnic/ && $4 == "HWaddr") {
				hwaddr[$1] = $5
				if (TRACE)
					print "hwaddr[" $1 "] = " hwaddr[$1]
			}
		}
	} else {
		while (("LC_ALL=C ifconfig" | getline) > 0) {
			if ($1 ~ /^vmnic/ && $4 == "HWaddr")
				hwaddr[$1] = $5
		}
	}

	# Walk through the vswif<n> and assign the FIRST found NIC
	# with the corresponding port group. Finally print association
	# as follows:
	#
	# FOR every vswif<n> DO
	# 	locate the portgroups[k,l] entry w.r.t. its portgroup value
	#	IF entry uplinks[k,l] exists THEN
	#		assign this entry as uplink
	#	ELSE
	#		assign the 1st entry uplinks[k,*] as uplink
	# DONE
	#	vswif0 00:50:56:42:5e:92 00:30:05:2F:A4:8C vmnic0
	#	vswif1 00:50:56:42:5e:93 00:30:05:2F:A4:8D vmnic1

	for (i = 0; i <= vswifNr; i++) {
		portgroup = vswifs[i, "PORTGROUP"]

		if (ESXVersion < "40") {

			for (j in portgroups) {
				if (portgroups[j] == portgroup) {
					if (j in uplinks)
						vswifs[i, "NIC"] = uplinks[j]
					else {
						split (j, idxj, SUBSEP)
						for (k in uplinks) {
							vswifs[i, "NIC"] = uplinks[k]
							split (k, idxk, SUBSEP)
							if (idxj[1] == idxk[1]  &&  idxj[2] < idxk[2])
								break
						}
					}

					if (TRACE)
						printf "found portgroup '%s' related with '%s'\n", portgroup, vswifs[i, "NIC"]
				}
			}
		
		} else {

			for (j in portgroups) {
				if (portgroups[j] == portgroup) {

					success = 0
					split (j, idxj, SUBSEP)

					for (k in PortGroupUplinks) {
						split (k, idxk, SUBSEP)

						if (idxj[1] == idxk[1] && idxj[2] == idxk[2]) {
							vswifs[i, "NIC"] = PortGroupUplinks[k]
							success = 1
							break
						}
					}

					if (success == 0) {
						for (k in VswitchUplinks) {
							split (k, idxk, SUBSEP)

							if (idxj[1] == idxk[1]) {
								vswifs[i, "NIC"] = VswitchUplinks[k]
								success = 1
								break
							}
						}
					}

					if (TRACE)
						printf "found portgroup '%s' related with '%s'\n", portgroup, vswifs[i, "NIC"]
				}
			}
		}

		# Add default field values for easier processing

		if (vswifs[i, "NIC"] == "")
			vswifs[i, "NIC"] = "unknown"
		if (hwaddr[vswifs[i, "NIC"]] == "")
			hwaddr[vswifs[i, "NIC"]] = "00:00:00:00:00:00"

		printf "%s %s %s %s\n",
			vswifs[i, "DEVICE"], vswifs[i, "MACADDR"],
			hwaddr[vswifs[i, "NIC"]], vswifs[i, "NIC"]

		if (TRACE == 2)
			printf "%s %s %s %s IPADDR: %s\n",
			vswifs[i, "DEVICE"], vswifs[i, "MACADDR"],
			hwaddr[vswifs[i, "NIC"]], vswifs[i, "NIC"],
			vswifs[i, "IPADDR"]
	}
}

# EOF
